==> scoras-0.3.2.tar.gz: scoras-0.3.2/scoras/rag.py <==
"""
RAG (Retrieval-Augmented Generation) implementation for Scoras.

This module provides classes and functions for implementing RAG systems,
which combine retrieval of relevant documents with generative models.
"""

import uuid
from typing import List, Dict, Any, Optional, Union, Callable

from .core import ScoringMixin

class Document:
    """
    Represents a document in a RAG system.
    
    A document is a piece of content that can be retrieved and used to augment
    the generation process.
    """
    
    def __init__(
        self,
        content: str,
        metadata: Optional[Dict[str, Any]] = None,
        id: Optional[str] = None
    ):
        """
        Initialize a document.
        
        Args:
            content: Content of the document
            metadata: Optional metadata for the document
            id: Optional unique identifier for the document
        """
        self.content = content
        self.metadata = metadata or {}
        self.id = id or str(uuid.uuid4())  # Generate a random ID if not provided
    
    def __str__(self) -> str:
        """
        Get a string representation of the document.
        
        Returns:
            String representation
        """
        return f"Document(id={self.id}, content={self.content[:50]}...)"
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the document to a dictionary.
        
        Returns:
            Dictionary representation
        """
        return {
            "id": self.id,
            "content": self.content,
            "metadata": self.metadata
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Document":
        """
        Create a document from a dictionary.
        
        Args:
            data: Dictionary representation
            
        Returns:
            Document instance
        """
        return cls(
            content=data["content"],
            metadata=data.get("metadata", {}),
            id=data.get("id")
        )

class SimpleRAG(ScoringMixin):
    """
    Implements a simple RAG system.
    
    This class provides a basic implementation of a RAG system that can be used
    to retrieve relevant documents and generate responses.
    """
    
    def __init__(
        self,
        agent: Any,
        documents: Optional[List[Document]] = None,
        enable_scoring: bool = True
    ):
        """
        Initialize a simple RAG system.
        
        Args:
            agent: Agent to use for generation
            documents: Optional list of documents
            enable_scoring: Whether to track complexity scoring
        """
        super().__init__(enable_scoring=enable_scoring)
        self.agent = agent
        self.documents = documents or []
        
        # Add RAG complexity score
        if self._enable_scoring:
            self._add_node_score("simple_rag", inputs=1, outputs=1)
            self._complexity_score.components["rag"] = 1.5  # RAG systems are more complex
            self._complexity_score.total_score += 1.5
            
            # Incorporate agent's complexity score
            if hasattr(agent, "get_complexity_score"):
                agent_score = agent.get_complexity_score()
                if isinstance(agent_score, dict) and "total_score" in agent_score:
                    self._complexity_score.total_score += agent_score["total_score"] * 0.5
                    self._complexity_score.components["agent"] = agent_score["total_score"] * 0.5
            
            self._complexity_score.update()
    
    def add_document(self, document: Document) -> None:
        """
        Add a document to the RAG system.
        
        Args:
            document: Document to add
        """
        self.documents.append(document)
        
        # Update complexity score
        if self._enable_scoring:
            self._add_edge_score(
                f"document_connection:{document.id}",
                path_distance=1.0,
                information_content=0.5
            )
            self._complexity_score.update()
    
    async def run(self, query: str) -> str:
        """
        Run the RAG system on the given query asynchronously.
        
        Args:
            query: Query to run
            
        Returns:
            Generated response
        """
        # In a real implementation, this would:
        # 1. Retrieve relevant documents based on the query
        # 2. Generate a prompt with the documents and query
        # 3. Call the agent with the prompt
        # For now, we'll just return a placeholder response
        return f"RAG response for: {query} (using {len(self.documents)} documents)"
    
    def run_sync(self, query: str) -> str:
        """
        Run the RAG system on the given query synchronously.
        
        Args:
            query: Query to run
            
        Returns:
            Generated response
        """
        # Synchronous version of run
        import asyncio
        return asyncio.run(self.run(query))

def create_document(
    content: str,
    metadata: Optional[Dict[str, Any]] = None,
    id: Optional[str] = None
) -> Document:
    """
    Create a document.
    
    Args:
        content: Content of the document
        metadata: Optional metadata for the document
        id: Optional unique identifier for the document
        
    Returns:
        Document instance
    """
    return Document(content=content, metadata=metadata, id=id)

def create_simple_rag(
    agent: Any,
    documents: Optional[List[Document]] = None
) -> SimpleRAG:
    """
    Create a simple RAG system.
    
    Args:
        agent: Agent to use for generation
        documents: Optional list of documents
        
    Returns:
        SimpleRAG instance
    """
    return SimpleRAG(agent=agent, documents=documents)

==> scoras-0.3.4.tar.gz: scoras-0.3.4/scoras/rag.py <==
"""
RAG (Retrieval-Augmented Generation) implementation for Scoras.

This module provides classes and functions for implementing RAG systems,
which combine retrieval of relevant documents with generative models.
"""

import uuid
from typing import List, Dict, Any, Optional, Union, Callable

from .core import ScoringMixin

class Document(BaseModel):
    """Model representing a document for retrieval."""
    
    content: str = Field(..., description="Content of the document")
    metadata: Dict[str, Any] = Field(default_factory=dict, description="Metadata associated with the document")
    embedding: Optional[List[float]] = Field(None, description="Vector embedding of the document")
    id: str = Field(default_factory=lambda: str(uuid.uuid4()), description="Unique identifier for the document")
    
    def __str__(self) -> str:
        """String representation of the document."""
        return self.content[:100] + "..." if len(self.content) > 100 else self.content

    
    def __str__(self) -> str:
        """
        Get a string representation of the document.
        
        Returns:
            String representation
        """
        return f"Document(id={self.id}, content={self.content[:50]}...)"
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the document to a dictionary.
        
        Returns:
            Dictionary representation
        """
        return {
            "id": self.id,
            "content": self.content,
            "metadata": self.metadata
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Document":
        """
        Create a document from a dictionary.
        
        Args:
            data: Dictionary representation
            
        Returns:
            Document instance
        """
        return cls(
            content=data["content"],
            metadata=data.get("metadata", {}),
            id=data.get("id")
        )

class SimpleRAG(ScoringMixin):
    """
    Implements a simple RAG system.
    
    This class provides a basic implementation of a RAG system that can be used
    to retrieve relevant documents and generate responses.
    """
    
    def __init__(
        self,
        agent: Any,
        documents: Optional[List[Document]] = None,
        enable_scoring: bool = True
    ):
        """
        Initialize a simple RAG system.
        
        Args:
            agent: Agent to use for generation
            documents: Optional list of documents
            enable_scoring: Whether to track complexity scoring
        """
        super().__init__(enable_scoring=enable_scoring)
        self.agent = agent
        self.documents = documents or []
        
        # Add RAG complexity score
        if self._enable_scoring:
            self._add_node_score("simple_rag", inputs=1, outputs=1)
            self._complexity_score.components["rag"] = 1.5  # RAG systems are more complex
            self._complexity_score.total_score += 1.5
            
            # Incorporate agent's complexity score
            if hasattr(agent, "get_complexity_score"):
                agent_score = agent.get_complexity_score()
                if isinstance(agent_score, dict) and "total_score" in agent_score:
                    self._complexity_score.total_score += agent_score["total_score"] * 0.5
                    self._complexity_score.components["agent"] = agent_score["total_score"] * 0.5
            
            self._complexity_score.update()
    
    def add_document(self, document: Document) -> None:
        """
        Add a document to the RAG system.
        
        Args:
            document: Document to add
        """
        self.documents.append(document)
        
        # Update complexity score
        if self._enable_scoring:
            self._add_edge_score(
                f"document_connection:{document.id}",
                path_distance=1.0,
                information_content=0.5
            )
            self._complexity_score.update()
    
    async def run(self, query: str) -> str:
        """
        Run the RAG system on the given query asynchronously.
        
        Args:
            query: Query to run
            
        Returns:
            Generated response
        """
        # In a real implementation, this would:
        # 1. Retrieve relevant documents based on the query
        # 2. Generate a prompt with the documents and query
        # 3. Call the agent with the prompt
        # For now, we'll just return a placeholder response
        return f"RAG response for: {query} (using {len(self.documents)} documents)"
    
    def run_sync(self, query: str) -> str:
        """
        Run the RAG system on the given query synchronously.
        
        Args:
            query: Query to run
            
        Returns:
            Generated response
        """
        # Synchronous version of run
        import asyncio
        return asyncio.run(self.run(query))

def create_document(
    content: str,
    metadata: Optional[Dict[str, Any]] = None,
    id: Optional[str] = None
) -> Document:
    """
    Create a document.
    
    Args:
        content: Content of the document
        metadata: Optional metadata for the document
        id: Optional unique identifier for the document
        
    Returns:
        Document instance
    """
    return Document(content=content, metadata=metadata, id=id)

def create_simple_rag(
    agent: Any,
    documents: Optional[List[Document]] = None
) -> SimpleRAG:
    """
    Create a simple RAG system.
    
    Args:
        agent: Agent to use for generation
        documents: Optional list of documents
        
    Returns:
        SimpleRAG instance
    """
    return SimpleRAG(agent=agent, documents=documents)
class ContextualRAG(SimpleRAG):
    """
    Extends SimpleRAG with contextual awareness.
    
    This class provides a more advanced RAG system that takes context into account
    when retrieving documents and generating responses.
    """
    
    def __init__(
        self,
        agent: Any,
        documents: Optional[List[Document]] = None,
        context_window: int = 5,
        enable_scoring: bool = True
    ):
        """
        Initialize a contextual RAG system.
        
        Args:
            agent: Agent to use for generation
            documents: Optional list of documents
            context_window: Size of the context window
            enable_scoring: Whether to track complexity scoring
        """
        super().__init__(agent=agent, documents=documents, enable_scoring=enable_scoring)
        self.context_window = context_window
        
        # Add additional complexity score for contextual awareness
        if self._enable_scoring:
            self._complexity_score.components["contextual"] = 1.0
            self._complexity_score.total_score += 1.0
            self._complexity_score.update()
    
    async def run(self, query: str, context: Optional[str] = None) -> str:
        """
        Run the contextual RAG system on the given query asynchronously.
        
        Args:
            query: Query to run
            context: Optional additional context
            
        Returns:
            Generated response
        """
        # In a real implementation, this would:
        # 1. Retrieve relevant documents based on the query and context
        # 2. Generate a prompt with the documents, query, and context
        # 3. Call the agent with the prompt
        # For now, we'll just return a placeholder response
        context_info = f" with context" if context else ""
        return f"Contextual RAG response for: {query}{context_info} (using {len(self.documents)} documents)"
    
    def run_sync(self, query: str, context: Optional[str] = None) -> str:
        """
        Run the contextual RAG system on the given query synchronously.
        
        Args:
            query: Query to run
            context: Optional additional context
            
        Returns:
            Generated response
        """
        # Synchronous version of run
        import asyncio
        return asyncio.run(self.run(query, context))
    
class Retriever(ScoringMixin):
    """
    Base class for document retrievers in a RAG system.
    
    A retriever is responsible for finding relevant documents based on a query.
    """
    
    def __init__(self, enable_scoring: bool = True):
        """
        Initialize a retriever.
        
        Args:
            enable_scoring: Whether to track complexity scoring
        """
        super().__init__(enable_scoring=enable_scoring)
        
        # Add retriever complexity score
        if self._enable_scoring:
            self._add_node_score("retriever", inputs=1, outputs=1)
            self._complexity_score.components["retrieval"] = 1.0
            self._complexity_score.total_score += 1.0
            self._complexity_score.update()
    
    async def retrieve(self, query: str, top_k: int = 5) -> List[Document]:
        """
        Retrieve relevant documents based on the query asynchronously.
        
        Args:
            query: Query to retrieve documents for
            top_k: Number of documents to retrieve
            
        Returns:
            List of retrieved documents
        """
        # This is a base class, subclasses should implement this method
        raise NotImplementedError("Subclasses must implement retrieve method")
    
    def retrieve_sync(self, query: str, top_k: int = 5) -> List[Document]:
        """
        Retrieve relevant documents based on the query synchronously.
        
        Args:
            query: Query to retrieve documents for
            top_k: Number of documents to retrieve
            
        Returns:
            List of retrieved documents
        """
        # Synchronous version of retrieve
        import asyncio
        return asyncio.run(self.retrieve(query, top_k))
